/* $Id: rc_imp.h,v 2.20 2017/05/03 19:34:21 hurtta Exp $ */

/******************************************************************************
 *  The Elm (ME+) Mail System  -  $Revision: 2.20 $   $State: Exp $
 *
 *  Author: Kari Hurtta <hurtta+elm@siilo.FMI.FI> 
 *                  (was hurtta+elm@posti.FMI.FI, hurtta+elm@ozone.FMI.FI)
 *      or  Kari Hurtta <elm@elmme-mailer.org>
 *****************************************************************************/

#if ANSI_C
struct rc_save_info_rec;
struct parsed_rc;
struct hist_record;
#endif

/* Returns 0 on error
           1 if parsed OK
*/
typedef int rc_parse_line P_((struct rc_save_info_rec *r,
			      int lcl,
			      struct string *value, int lineno, 
			      const char *filename,
			      int negate));
typedef int rc_parse_cline P_((struct rc_save_info_rec *r,
			       int lcl,
			       struct string *value, int lineno, 
			       const char *filename));

typedef void rc_print_value P_((FILE *F,
				struct rc_save_info_rec *r,
				int comment,
				charset_t rcset));

/* Returns pointer to static buffer */
typedef char * rc_get_value P_((struct rc_save_info_rec *r));

/* frees memory alloced by option */
typedef void rc_free_option  P_((struct rc_save_info_rec *r));




#define RCTYPE_magic            0xF402

extern struct rc_type {
    unsigned short magic;                      /* RCTYPE_magic */


    rc_parse_line      * parse_line;            /* Parse line */
    rc_parse_cline     * parse_cline;           /* Parse continuation line */
    rc_print_value     * print_value;
    rc_get_value       * get_value;
    rc_free_option     * free_option;

} rc_DT_SYN, rc_DT_STR, rc_DT_NUM, rc_DT_BOL,
    rc_DT_CHR, rc_DT_SORT,
    rc_DT_MLT,  rc_DT_PRM, rc_DT_FUNC,
    rc_DT_LONG, rc_DT_PATH, rc_DT_OBS, rc_DT_ESTR,
    rc_DT_ENUM, rc_DT_DELAY, rc_DT_STRING,
    rc_DT_FLAGS;

extern int valid_rc_type P_((struct rc_type *T));

extern struct rc_type rc_DT_PGPVER;

#ifdef USE_DLOPEN
extern struct rc_type rc_DT_SHARED;
#endif

#define DT_SYN      &rc_DT_SYN    /* synonym entry (old name) */
#define DT_STR      &rc_DT_STR    /* string */
#define DT_NUM      &rc_DT_NUM    /* number */
#define DT_BOL      &rc_DT_BOL    /* ON/OFF (boolean) */
#define DT_WEE      &rc_DT_WEE    /* weed list */
#define DT_MLT      &rc_DT_MLT    /* multiple destinations for data */
#define DT_PRM      &rc_DT_PRM   /* file permissions */
#define DT_FUNC     &rc_DT_FUNC  /* function to be called for process option */
#define DT_LONG     &rc_DT_LONG    /* number */
#define DT_PGPVER   &rc_DT_PGPVER    /* enum pgp_version */

/* read_rc.c. -- return NULL on failure 
                 otherwise pointer to option
*/
extern struct rc_save_info_rec  * do_set P_((int lcl, char *word1, 
					     struct string *word2, 
					     const char *filename));
extern void weedout P_((char *string)); 

extern int expand_dt_path P_((struct dt_path_info *ptr, const char *fieldname, 
			      const char *val, int recursive));

extern int sconvert_to_rcset P_((struct string *S,const char *fieldname,
				  charset_t rcset,
				  char **buffer,
				  int *buffer_len));

/* Return 1 if no errors */
extern int convert_to_system_charset P_((struct string *value,
					 int lineno,
					 const char *filename,
					 char **buffer,
					 int *buffer_len));

/* in_buffer_len = -1 ... uses  strlen() 
   0 == failure, 1 = ok
*/

extern int convert_to_rcset P_((const char * in_buffer,int in_buffer_len,
				const char *fieldname,
				charset_t rcset,
				char **buffer,
				int *buffer_len));
	
/* Expands ptr->unexpanded 
   0 = failed to expand
   1 = expanded
   -1 = expanded but do string available
*/
extern int expand_dt_estr P_((struct dt_estr_info *ptr,
			      const char *fieldname));

extern void estr_set_expanded P_((struct dt_estr_info *ptr, 
				  const char *buffer));


extern void mark_XX P_((struct rc_save_info_rec *x,
			int no_history_record));

#ifdef USE_DLOPEN
extern void mark_shared_changed P_((void *A,
				    int no_history_record));
extern void mark_fshared_changed P_((option_func *A,
				     int no_history_record));
extern struct dt_enum_shared *  
   shared_fill_dt_enumerate_values P_((struct dt_enumerate_info *ptr,
				       const char *tag));
#endif

extern void register_delayed P_((struct rc_save_info_rec * rc_options,
				 int rc_option_count));
extern int rc_eval_tail P_((struct rc_save_info_rec *ret,int lcl, 
			    struct string *value,
			    int lineno, const char *filename, int negate));

extern void  mark_delayed_changed P_((void *A, int no_history_record));
extern void  mark_fdelayed_changed P_((option_func *A,
				       int no_history_record));

extern void free_delayed_rc_options P_((void));

extern charset_t rcset;             /* What charset use for writing */
extern struct editor_propline * current_editor_propline; /* Also other values */

extern void free_editor_propline P_((struct editor_propline **propline));
extern struct editor_propline *detect_editor_propline 
  P_((enum editor_propline_v detect_mode,
      char * buffer, int buffer_len,
      const char *filename, int lineno,
      charset_t *current_charset));

struct editor_propline * copy_editor_propline P_((const struct editor_propline 
						  *propline));


extern void write_editor_propline P_((const struct editor_propline *propline,
				      enum editor_propline_v detect_mode,
				      FILE *f,
				      const char * line_prefix, 
				      const char * line_tail,
				      charset_t file_charset));


extern void free_parsed_rc P_((struct parsed_rc **rc));

#define MAX_LINE 10000

extern struct parsed_rc * do_rc_parse P_((FILE *file,
					  enum record_mode lcl,
					  const char *filename,
					  int *errors_p));

extern void do_rc_process P_((struct parsed_rc *rc, enum record_mode lcl,
			      const char *filename, charset_t elmrc_charset,
			      int *errors_p));

extern void hist_record_add_line P_((struct hist_record *rec, char *rest));
extern void hist_record_first_line P_((struct hist_record *rec, char *rest));
extern void hist_record_changed P_((struct hist_record *rec));

extern char * rc_split_keyword P_((char *buffer, int line_len,
				   char *keyword, int size,
				   char **tag,
				   int *assig,
				   int *restlen));

extern struct rc_save_info_rec *rc_locate_option P_((struct rc_save_info_rec * 
						     rc_options,
						     int rc_option_count,
						     char *word1,
						     int *negate));

extern struct hist_record * rc_need_hist_record P_((enum record_mode mode,
						    struct rc_save_info_rec *X,
						    char *tag));


/* service_list.c */

/* Mallocs to result */
extern char * convert_to_fileset2 P_((const struct string *value, 
				      charset_t fileset, int *err));

/* Mallocs to result, if conversion, otherwise returns value */
extern char * convert_to_fileset P_((char *value, charset_t fileset, int *err));

/*
 * Local Variables:
 *  mode:c
 *  c-basic-offset:4
 *  buffer-file-coding-system: iso-8859-1
 * End:
 */
